import 'dart:convert';

import 'package:models_weebi/abstracts.dart';
import 'package:collection/collection.dart';
import 'package:models_weebi/src/models/item_in_cart/item_cart_abstract.dart';
import 'package:models_weebi/models.dart';

// items are deserialized in tickets with a single generic
// since a ticket can contain item with articleW and articleBasket
// we keep a single class itemWeebi to handle it

extension AggregateItems on Iterable<ItemCartWeebi> {
  num get itemsTotalPrice => fold(0, (value, item) {
        if (item.isBasket) {
          return (value +
                  (item.proxiesWorth as Iterable<ProxyArticleWorth>)
                      .basketTotalPrice(
                          (item.article as ArticleBasketOnTicket)
                              .discountAmountSalesOnly,
                          (item.article as ArticleBasketOnTicket)
                              .markupAmountSalesOnly)) *
              item.quantity;
        } else {
          return (value +
              ((item.article as ArticleRetailOnTicket).price * item.quantity));
        }
      });

  // -----
  // below spend and spendDeferred
  // -----
  num get itemsTotalCost => fold(0, (value, item) {
        if (item.isBasket) {
          return value +
              (item.proxiesWorth as Iterable<ProxyArticleWorth>).baskeTotalCost(
                      0, //(item.article as ArticleBasketOnTicket).discountAmountSalesOnly,
                      0 //(item.article as ArticleBasketOnTicket).markupAmountSalesOnly)
                      ) *
                  item.quantity;
          // here I do not apply discount since this is a purchase
          // if needed consider creating an additional field discountAmountPurchaseOnly
        } else {
          return value +
              ((item.article as ArticleRetailOnTicket).cost * item.quantity);
        }
      });
}

class ItemCartWeebi<A extends ArticleRawAbstract> extends ItemAbstract<A> {
  /// used only when using the inventory functionnality
  final double? inventoryAbsoluteQt;

  //
  ItemCartWeebi(
    super.articleCreator,
    super.quantity, {
    super.proxiesWorth,
    this.inventoryAbsoluteQt,
  });

  @override
  String toString() =>
      "Item($article,$quantity,$proxiesWorth,$inventoryAbsoluteQt)";

  @override
  Map<String, dynamic> toMap() {
    return {
      'article': article.toMap(),
      'quantity': quantity,
      'proxies_worth': proxiesWorth?.map((x) => x.toMap()).toList(),
      'inventoryAbsoluteQt': inventoryAbsoluteQt
    };
  }

  @override
  String toJson() => json.encode(toMap());

  bool get isBasket => toMap()['article']['proxies'] != null;

  static final dummy =
      ItemCartWeebi(() => ArticleRetailOnTicket.dummyRetailOnTicket, 1.0);

  static final dummyBasket = ItemCartWeebi(
      () => ArticleBasketOnTicket.dummyBasketOnTicket, 1.0,
      proxiesWorth: <ProxyArticleWorth>[ProxyArticleWorth.dummy]);

  // I can play around with Article while ignoring its exact type :)
  num get totalPrice => (articlePrice * quantity);

  num get totalCost => (articleCost * quantity);

  num get articlePrice {
    A article = articleCreator();
    if (article.toMap()['proxies'] != null) {
      return proxiesWorth?.basketTotalPrice(
              (article as ArticleBasketOnTicket).discountAmountSalesOnly,
              (article as ArticleBasketOnTicket).markupAmountSalesOnly) ??
          0;
    } else {
      return (article as ArticleRetailOnTicket).price;
    }
  }

  num get articleCost {
    A article = articleCreator();
    // no basket purchase today 2024 April
    // if (article.toMap()['proxies'] != null) {
    // return proxiesWorth?.baskeTotalCost(0, 0) ?? 0;
    // * not applying any discount in purchase, discount& markup ar for sales only
    //(article as ArticleBasketOnBasket).discountAmountSalesOnly,
    //(article as ArticleBasketOnBasket).markupAmountSalesOnly)
    // } else {
    return (article as ArticleRetailOnTicket).cost;
    // }
  }

  double getStockMovementForArticle<AA extends ArticleRawAbstract>(
      AA aSelected) {
    double stockMovement = 0.0;
    A aInItem = articleCreator();
    if (aSelected.calibreId == aInItem.calibreId &&
        aSelected.id == aInItem.id) {
      stockMovement += quantity; // weight should not be included in article
    } else {
      if (isBasket == true) {
        if ((aInItem as ArticleBasketOnTicket).proxies.any((proxy) =>
            proxy.proxyCalibreId == aSelected.calibreId &&
            proxy.proxyArticleId == aSelected.id)) {
          for (final proxyOfArticleInItem in aInItem.proxies.where((proxy) =>
              proxy.proxyCalibreId == aSelected.calibreId &&
              proxy.proxyArticleId == aSelected.id)) {
            stockMovement += quantity *
                proxyOfArticleInItem
                    .minimumUnitPerBasket; // weight should not be included in article
          }
        }
      }
    }
    return stockMovement;
  }

  double getStockMovementForCalibre(Calibre calibre) {
    double stockMovement = 0.0;
    A aInItem = articleCreator();
    if (calibre.id == aInItem.calibreId && calibre.isBasket == false) {
      stockMovement += quantity *
          (aInItem as ArticleRetailOnTicket)
              .unitsInOnePiece; // weight needed for calibre full stock
    } else {
      if (isBasket == true) {
        if ((aInItem as ArticleBasketOnTicket)
            .proxies
            .any((p) => p.proxyCalibreId == calibre.id)) {
          for (final proxyOfArticleInItem
              in aInItem.proxies.where((p) => p.proxyCalibreId == calibre.id)) {
            stockMovement += quantity *
                proxyOfArticleInItem.minimumUnitPerBasket *
                proxyOfArticleInItem
                    .articleWeight; // weight needed for line full stock
          }
        }
      }
    }
    return stockMovement;
  }

  factory ItemCartWeebi.fromJson(String source) =>
      ItemCartWeebi.fromMap(json.decode(source));

  static fromMap(Map<String, dynamic> map) {
    return map['article']['proxies'] == null
        ? fromMapRetail(map)
        : fromMapBasket(map);
  }

  static ItemCartWeebi<ArticleBasketOnTicket> fromMapBasket(
      Map<String, dynamic> map) {
    return ItemCartWeebi<ArticleBasketOnTicket>(
      () => ArticleBasketOnTicket.fromMap(map['article']),
      map['quantity'] == null ? 0.0 : (map['quantity'] as num).toDouble(),
      proxiesWorth: map['proxies_worth'] != null
          ? List<ProxyArticleWorth>.from(
              map['proxies_worth']?.map((x) => ProxyArticleWorth.fromMap(x)))
          : <ProxyArticleWorth>[],
    );
  }

  static ItemCartWeebi<ArticleRetailOnTicket> fromMapRetail(
      Map<String, dynamic> map) {
    return ItemCartWeebi<ArticleRetailOnTicket>(
      () => ArticleRetailOnTicket.fromMap(map['article']),
      map['quantity'] == null ? 0.0 : (map['quantity'] as num).toDouble(),
      proxiesWorth: [],
      inventoryAbsoluteQt: map['inventoryAbsoluteQt'] == null
          ? null
          : (map['inventoryAbsoluteQt'] as num).toDouble(),
    );
  }

  ItemCartWeebi copyWith(
      {ArticleCreator<A>? articleCreator,
      double? quantity,
      List<ProxyArticleWorth>? proxiesWorth,
      double? inventoryAbsoluteQt}) {
    return ItemCartWeebi<A>(
      articleCreator ?? this.articleCreator,
      quantity ?? this.quantity,
      proxiesWorth: proxiesWorth ?? this.proxiesWorth,
      inventoryAbsoluteQt: inventoryAbsoluteQt ?? this.inventoryAbsoluteQt,
    );
  }

  @override
  bool operator ==(Object other) {
    if (identical(this, other)) return true;
    final listEquals = const DeepCollectionEquality().equals;
    return other is ItemCartWeebi &&
        other.quantity == quantity &&
        other.article == article &&
        listEquals(other.proxiesWorth, proxiesWorth);
  }

  @override
  int get hashCode => quantity.hashCode ^ article.hashCode;
}
